import logging

from gongish.router import RouterMixin
from gongish.configuration import ConfigurationMixin


logger = logging.getLogger("gongish")


class Application(RouterMixin, ConfigurationMixin):

    #: Application logger based on python builtin logging module
    __logger__ = logger

    def __init__(self):
        ConfigurationMixin.__init__(self)
        RouterMixin.__init__(self)

    def setup(self):  # pragma: nocover
        raise NotImplementedError

    def shutdown(self):  # pragma: nocover
        raise NotImplementedError



import sys
import inspect
import functools

from .request import Request
from .response import Response
from .exceptions import (
    HTTPStatus,
    HTTPSuccess,
    HTTPNotFound,
    HTTPBadRequest,
    HTTPInternalServerError,
)
from .response_formatters import ResponseFormattersMixin
from .static import StaticHandlerMixin
from .helpers import WordRouter

http_success = HTTPSuccess().status


class RouterMixin(StaticHandlerMixin, ResponseFormattersMixin):
    __request_factory__ = Request
    __response_factory__ = Response
    __route_argument_char__ = ":"

    def __init__(self):
        self.request = None
        self.response = None
        self.verbs = set()
        self.paths = set()
        self.wordrouter = WordRouter()

    def on_begin_request(self):
        """ Hook """
        pass

    def on_begin_response(self):
        """ Hook """
        pass

    def on_end_response(self):
        """ Hook """
        pass

    def on_error(self, exc: HTTPStatus):
        """ Hook """
        pass

    def __getattr__(self, key):
        formatter = super().__getattribute__(f"format_{key}")
        return functools.partial(self.route, formatter=formatter)

    def chunked(self, trailer_field=None, trailer_value=None):
        """
        http://tools.ietf.org/html/rfc2616#section-14.40
        http://tools.ietf.org/html/rfc2616#section-14.41
        """
        app = self

        def decorator(func):
            @functools.wraps(func)
            def wrapper(*args, **kwargs):
                nonlocal trailer_field, trailer_value, app
                app.response.headers["transfer-encoding"] = "chunked"
                if trailer_field:
                    app.response.headers["trailer"] = trailer_field

                result = func(*args, **kwargs)
                try:
                    while True:
                        chunk = next(result)
                        yield f"{len(chunk)}\r\n{chunk}\r\n"

                except StopIteration:
                    yield "0\r\n"
                    if trailer_field and trailer_value:
                        yield f"{trailer_field}: {trailer_value}\r\n"
                    yield "\r\n"

                except Exception as ex:
                    exstr = str(ex)
                    yield f"{len(exstr)}\r\n{exstr}"
                    yield "0\r\n\r\n"

            return wrapper

        if callable(trailer_field):
            func = trailer_field
            trailer_field = None
            return decorator(func)

        return decorator

    def route(self, path, formatter=None, **kwargs):
        def decorator(func):
            funcname = func.__name__.lower()

            # Set formatter
            func.__gongish_formatter__ = functools.partial(
                formatter or self.__class__.__default_formatter__, **kwargs
            )

            # Get parameters
            func.__gongish_route_params__ = tuple(
                [
                    (param.name, param.annotation)
                    for param in inspect.signature(func).parameters.values()
                ]
            )

            # Distribute
            self.wordrouter.add(funcname + path, func)
            self.paths.add(path)
            self.verbs.add(funcname)

        return decorator

    def handle_exception(self, exc, start_response):
        if isinstance(exc, HTTPStatus):
            exc_ = exc
            exc_info = None

        else:
            exc_ = HTTPInternalServerError()
            exc_info = sys.exc_info()
            self.__logger__.exception(exc_.text, exc_info=True)

        exc_.setup_response(app=self)

        self.on_error(exc_)  # hook

        self.response.prepare_to_start()

        start_response(
            exc_.status, list(self.response.headers.items()), exc_info
        )

        return self.response.start()

    def dispatch(self, path: str, verb: str):
        """
        Dispatch path and return handler function
        """

        handler, route_args = self.wordrouter.dispatch(verb + path)
        if not handler:
            raise HTTPNotFound

        # Validate parameters
        for idx, param in enumerate(handler.__gongish_route_params__):
            name, annotation = param

            if annotation == inspect._empty:
                continue

            # Call the annotation
            try:
                route_args[idx] = annotation(route_args[idx])

            except Exception:
                raise HTTPBadRequest(f"Invalid Parameter `{name}`")

        return handler, route_args

    def __call__(self, environ, start_response):
        """ Application WSGI entry """
        self.request = request = self.__request_factory__(environ)
        self.response = response = self.__response_factory__(app=self)
        try:
            self.on_begin_request()  # hook

            # Call handler
            handler, route_args = self.dispatch(request.path, request.verb)
            response.body = handler(*route_args)
            response.status = http_success

            # Format response
            handler.__gongish_formatter__(request, response)

            response.prepare_to_start()

        except Exception as exc:
            return self.handle_exception(exc, start_response)

        self.on_begin_response()  # hook

        # Cookies
        cookie = request.cookies.output()
        if cookie:
            for line in cookie.split("\r\n"):
                response.headers.add(*line.split(": ", 1))

        start_response(response.status, list(response.headers.items()))

        return response.start()



import sys

from . import serve


def main():
    if len(sys.argv) > 0 and sys.argv[1] == "serve":
        return serve.main(sys.argv[2:])

    print("\n".join(("Gongish CLI", "", "Commands:", "gongish serve")))



import traceback

from gongish.helpers import HeaderSet


class HTTPStatus(Exception):
    code = None
    text = None
    __keep_headers__ = False
    __keep_body__ = True

    def __init__(self, *args: object) -> None:
        self.headers = HeaderSet()
        super().__init__(*args)

    @property
    def status(self):
        return f"{self.code} {self.text}"

    def setup_response(self, app) -> str:
        if not self.__class__.__keep_body__:
            app.response.body = None
            app.response.type = None
        else:
            app.response.charset = "utf-8"
            app.response.type = "text/plain"
            app.response.body = (
                traceback.format_exc() if app.config.debug else self.status
            )

        if self.__keep_headers__:
            app.response.headers.update(self.headers)
        else:
            app.response.headers = self.headers


class HTTPKnownStatus(HTTPStatus):
    def __init__(self, status_text=None):
        if status_text is not None:
            self.text = status_text
        super().__init__()


class HTTPBadRequest(HTTPKnownStatus):
    code = 400
    text = "Bad Request"


class HTTPUnauthorized(HTTPKnownStatus):
    code = 401
    text = "Unauthorized"


class HTTPForbidden(HTTPKnownStatus):
    code = 403
    text = "Forbidden"


class HTTPNotFound(HTTPKnownStatus):
    code = 404
    text = "Not Found"


class HTTPMethodNotAllowed(HTTPKnownStatus):
    code = 405
    text = "Method Not Allowed"


class HTTPConflict(HTTPKnownStatus):
    code = 409
    text = "Conflict"


class HTTPGone(HTTPKnownStatus):
    code = 410
    text = "Gone"


class HTTPPreconditionFailed(HTTPKnownStatus):
    code = 412
    text = "Precondition Failed"


class HTTPRedirect(HTTPKnownStatus):
    """
    This is an abstract class for all redirects.
    """

    def __init__(self, location, *args, **kw):
        super().__init__(*args, **kw)
        self.headers["Location"] = location


class HTTPMovedPermanently(HTTPRedirect):
    code = 301
    text = "Moved Permanently"
    __keep_body__ = False


class HTTPFound(HTTPRedirect):
    code = 302
    text = "Found"
    __keep_body__ = False


class HTTPNotModified(HTTPKnownStatus):
    code = 304
    text = "Not Modified"
    __keep_body__ = False


class HTTPInternalServerError(HTTPKnownStatus):
    code = 500
    text = "Internal Server Error"


class HTTPBadGatewayError(HTTPKnownStatus):
    code = 502
    text = "Bad Gateway"


class HTTPSuccess(HTTPKnownStatus):
    code = 200
    text = "OK"
    __keep_headers__ = True


class HTTPCreated(HTTPSuccess):
    code = 201
    text = "Created"
    __keep_headers__ = True


class HTTPAccepted(HTTPSuccess):
    code = 202
    text = "Accepted"
    __keep_headers__ = True


class HTTPNonAuthoritativeInformation(HTTPSuccess):
    code = 203
    text = "Non-Authoritative Information"
    __keep_headers__ = True


class HTTPNoContent(HTTPSuccess):
    code = 204
    text = "No Content"
    __keep_body__ = False
    __keep_headers__ = True


class HTTPResetContent(HTTPSuccess):
    code = 205
    text = "Reset Content"
    __keep_body__ = False
    __keep_headers__ = True


class HTTPPartialContent(HTTPSuccess):
    code = 206
    text = "Partial Content"
    __keep_headers__ = True



from setuptools import setup, find_packages


def read_version(module_name):
    from re import match, S
    from os.path import join, dirname

    with open(join(dirname(__file__), module_name, "__init__.py")) as f:
        return match(r".*__version__.*('|\")(.*?)('|\")", f.read(), S).group(2)


setup(
    name="gongish",
    version=read_version("gongish"),
    description="A simple and fast HTTP framework for Python",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="http://github.com/meyt/gongish",
    author="Mahdi Ghane.g",
    license="MIT",
    keywords="web tool-chain",
    packages=find_packages(),
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Unix",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Topic :: Software Development :: Libraries",
    ],
    entry_points={"console_scripts": ["gongish = gongish.cli:main"]},
    install_requires=["pymlconf == 2.2.0"],
)
